from __future__ import annotations

from abc import abstractmethod, ABC
from typing import cast
from weakref import WeakValueDictionary

from sapiopylib.rest.User import SapioUser

from sapiopycommons.files.file_bridge import FileBridge, FileBridgeMetadata
from sapiopycommons.general.aliases import AliasUtil, UserIdentifier


class FileBridgeHandler:
    """
    The FileBridgeHandler provides caching of the results of file bridge endpoint calls while also containing quality
    of life functions for common file bridge actions.
    """
    user: SapioUser
    __bridge: str
    __file_data_cache: dict[str, bytes]
    """A cache of file paths to file bytes."""
    __file_objects: dict[str, File]
    """A cache of file paths to File objects."""
    __dir_file_name_cache: dict[str, list[str]]
    """A cache of directory file paths to the names of the files or nested directories within them."""
    __dir_objects: dict[str, Directory]
    """A cache of directory file paths to Directory objects."""
    __file_metadata_cache: dict[str, FileBridgeMetadata]
    """A cache of file or directory paths to file metadata."""
    __dir_metadata_cache: dict[str, list[FileBridgeMetadata]]
    """A cache of directory file paths to the metadata of the files or nested directories within them."""

    __instances: WeakValueDictionary[str, FileBridgeHandler] = WeakValueDictionary()
    __initialized: bool

    def __new__(cls, context: UserIdentifier, bridge_name: str):
        """
        :param context: The current webhook context or a user object to send requests from.
        """
        user = AliasUtil.to_sapio_user(context)
        key = f"{user.__hash__()}:{bridge_name}"
        obj = cls.__instances.get(key)
        if not obj:
            obj = object.__new__(cls)
            obj.__initialized = False
            cls.__instances[key] = obj
        return obj

    def __init__(self, context: UserIdentifier, bridge_name: str):
        """
        :param context: The current webhook context or a user object to send requests from.
        :param bridge_name: The name of the bridge to communicate with. This is the "connection name" in the
            file bridge configurations.
        """
        if self.__initialized:
            return
        self.__initialized = True

        self.user = AliasUtil.to_sapio_user(context)
        self.__bridge = bridge_name
        self.__file_data_cache = {}
        self.__file_objects = {}
        self.__dir_file_name_cache = {}
        self.__dir_objects = {}
        self.__file_metadata_cache = {}
        self.__dir_metadata_cache = {}

    @property
    def connection_name(self) -> str:
        return self.__bridge

    def clear_caches(self) -> None:
        """
        Clear the file and directory caches of this handler.
        """
        self.__file_data_cache.clear()
        self.__file_objects.clear()
        self.__dir_file_name_cache.clear()
        self.__dir_objects.clear()
        self.__file_metadata_cache.clear()
        self.__dir_metadata_cache.clear()

    # CR-47388: Allow the FileBridgeHandler's File and Directory objects to be provided as file path parameters.
    def file_exists(self, file_path: str | File | Directory) -> bool:
        """
        Determine if a file or directory exists in FileBridge at the provided path. This is achieved by calling for the
        metadata of the provided file path. If the file does not exist, then an exception is raised, which is caught and
        handled by this function as a return value of False.

        :param file_path: A file path, File object, or Directory object.
        :return: True if the file exists. False if it does not.
        """
        if isinstance(file_path, FileBridgeObject):
            file_path = file_path.path
        try:
            self.file_metadata(file_path)
            return True
        except Exception:
            return False

    def read_file(self, file_path: str | File, base64_decode: bool = True) -> bytes:
        """
        Read a file from FileBridge. The bytes of the given file will be cached so that any subsequent reads of this
        file will not make an additional webservice call.

        :param file_path: The file path or File object to read from.
        :param base64_decode: If true, base64 decode the file. Files are base64 encoded by default when retrieved from
            FileBridge.
        :return: The bytes of the file.
        """
        if isinstance(file_path, File):
            file_path = file_path.path
        if file_path in self.__file_data_cache:
            return self.__file_data_cache[file_path]
        file_bytes: bytes = FileBridge.read_file(self.user, self.__bridge, file_path, base64_decode)
        self.__file_data_cache[file_path] = file_bytes
        return file_bytes

    def write_file(self, file_path: str | File, file_data: bytes | str) -> None:
        """
        Write a file to FileBridge. The bytes of the given file will be cached so that any subsequent reads of this
        file will not make an additional webservice call.

        :param file_path: The file path or File object to write to. If a file already exists at the given path then the
            file is overwritten.
        :param file_data: A string or bytes of the file to be written.
        """
        if isinstance(file_path, File):
            file_path = file_path.path
        FileBridge.write_file(self.user, self.__bridge, file_path, file_data)
        self.__file_data_cache[file_path] = file_data if isinstance(file_data, bytes) else file_data.encode()

        # Find the directory path to this file and the name of the file. Add the file name to the cached list of
        # files for the directory, assuming we have this directory cached and the file isn't already in it.
        file_name, path_to = split_path(file_path)
        if path_to in self.__dir_file_name_cache and file_name not in self.__dir_file_name_cache[path_to]:
            self.__dir_file_name_cache[path_to].append(file_name)

    def delete_file(self, file_path: str | File) -> None:
        """
        Delete an existing file in FileBridge. If this file is in the cache, it will also be deleted from the cache.

        :param file_path: The file path or File object to delete.
        """
        if isinstance(file_path, File):
            file_path = file_path.path
        FileBridge.delete_file(self.user, self.__bridge, file_path)
        if file_path in self.__file_data_cache:
            self.__file_data_cache.pop(file_path)
        if file_path in self.__file_objects:
            self.__file_objects.pop(file_path)
        if file_path in self.__file_metadata_cache:
            self.__file_metadata_cache.pop(file_path)

    def list_directory(self, file_path: str | Directory) -> list[str]:
        """
        List the contents of a FileBridge directory. The contents of this directory will be cached so that any
        subsequent lists of this directory will not make an additional webservice call.

        :param file_path: The directory path or Directory object to list from.
        :return: A list of names of files and folders in the directory.
        """
        if isinstance(file_path, Directory):
            file_path = file_path.path
        if file_path in self.__dir_file_name_cache:
            return self.__dir_file_name_cache[file_path]
        files: list[str] = FileBridge.list_directory(self.user, self.__bridge, file_path)
        self.__dir_file_name_cache[file_path] = files
        return files

    def create_directory(self, file_path: str | Directory) -> None:
        """
        Create a new directory in FileBridge. This new directory will be added to the cache as empty so that listing
        the same directory does not make an additional webservice call.

        :param file_path: The directory path or Directory object to create the directory at. If a directory already
            exists at the given path then an exception is raised.
        """
        if isinstance(file_path, Directory):
            file_path = file_path.path
        FileBridge.create_directory(self.user, self.__bridge, file_path)
        # This directory was just created, so we know it's empty.
        self.__dir_file_name_cache[file_path] = []
        self.__dir_metadata_cache[file_path] = []

        # Find the directory path to this directory and the name of the directory. Add the directory name to the cached
        # list of files for the directory, assuming we have this directory cached and the directory isn't already in it.
        dir_name, path_to = split_path(file_path)
        if path_to in self.__dir_file_name_cache and dir_name not in self.__dir_file_name_cache[path_to]:
            self.__dir_file_name_cache[path_to].append(dir_name)

    def delete_directory(self, file_path: str | Directory) -> None:
        """
        Delete an existing directory in FileBridge. If this directory is in the cache, it will also be deleted
        from the cache.

        :param file_path: The directory path or Directory object to delete.
        """
        if isinstance(file_path, Directory):
            file_path = file_path.path
        FileBridge.delete_directory(self.user, self.__bridge, file_path)
        # The deletion of a directory also deletes all the files within it, so we need to check every cache for file
        # paths that start with this directory path and remove them.
        for key in list(self.__file_data_cache.keys()):
            if key.startswith(file_path):
                self.__file_data_cache.pop(key)
        for key in list(self.__file_objects.keys()):
            if key.startswith(file_path):
                self.__file_objects.pop(key)
        for key in list(self.__file_metadata_cache.keys()):
            if key.startswith(file_path):
                self.__file_metadata_cache.pop(key)
        for key in list(self.__dir_file_name_cache.keys()):
            if key.startswith(file_path):
                self.__dir_file_name_cache.pop(key)
        for key in list(self.__dir_objects.keys()):
            if key.startswith(file_path):
                self.__dir_objects.pop(key)
        for key in list(self.__dir_metadata_cache.keys()):
            if key.startswith(file_path):
                self.__dir_metadata_cache.pop(key)

    # FR-47387: Add support for the metadata endpoints in FileBridge.
    def file_metadata(self, file_path: str | File | Directory) -> FileBridgeMetadata:
        """
        Get metadata for a file in FileBridge. If this metadata is already cached, then it will be returned from the
        cache.

        The file path may be to a directory, in which case only the metadata for that directory will be returned. If you
        want the metadata for the contents of a directory, then use the directory_metadata function.

        :param file_path: The file path, File object, or Directory object to retrieve the metadata from.
        :return: The metadata for the file.
        """
        if isinstance(file_path, FileBridgeObject):
            file_path = file_path.path
        if file_path in self.__file_metadata_cache:
            return self.__file_metadata_cache[file_path]
        metadata: FileBridgeMetadata = FileBridge.file_metadata(self.user, self.__bridge, file_path)
        self.__file_metadata_cache[file_path] = metadata

        # It's possible that this file is newly created, but the directory it's in was already cached. If that's the
        # case, then we need to add this file's metadata to the directory's cache. (The write_file/create_directory
        # methods will have already handled the directory's file name cache.)
        file_name, path_to = split_path(file_path)
        if (path_to in self.__dir_metadata_cache
                and not any([file_name == x.file_name for x in self.__dir_metadata_cache[path_to]])):
            self.__dir_metadata_cache[path_to].append(metadata)
        return metadata

    def directory_metadata(self, file_path: str | Directory) -> list[FileBridgeMetadata]:
        """
        Get metadata for every file in a directory in FileBridge. If this metadata is already cached, then it will be
        returned from the cache.

        :param file_path: The path to the directory to retrieve the metadata of the contents, or the Directory object.
        :return: A list of the metadata for each file in the directory.
        """
        if isinstance(file_path, Directory):
            file_path = file_path.path
        # If the directory metadata is already cached, then use the cached value instead of making an additional
        # webservice call. The only exception to this is if the size of the directory's file name cache differs from
        # the size of the directory's metadata cache. This can happen if a new file or directory was added to the
        # directory using write_file/create_directory after the metadata of the directory's contents was cached.
        # In this case, we need to make an additional webservice call to get the metadata of the new file or directory.
        # Since there could be multiple new files or directories, just re-query the metadata for the entire directory.
        if (file_path in self.__dir_metadata_cache
                and len(self.__dir_metadata_cache[file_path]) == len(self.__dir_file_name_cache[file_path])):
            return self.__dir_metadata_cache[file_path]
        metadata: list[FileBridgeMetadata] = FileBridge.directory_metadata(self.user, self.__bridge, file_path)
        # Save the metadata to the directory cache.
        self.__dir_metadata_cache[file_path] = metadata
        # We can also save the metadata to the file cache so that we don't have to make additional webservice calls if
        # an individual file's metadata is requested.
        for file_metadata in metadata:
            self.__file_metadata_cache[file_path + "/" + file_metadata.file_name] = file_metadata
        # This also doubles as a list directory call since it contains the file names of the contents of the directory.
        self.__dir_file_name_cache[file_path] = [x.file_name for x in metadata]
        return metadata

    def is_file(self, file_path: str | File | Directory) -> bool:
        """
        Determine if the given file path points to a file. This is achieved by checking the metadata of the provided
        file path. If the metadata is not cached, then this will make a webservice call to get the metadata.

        :param file_path: A file path, File object, or Directory object.
        :return: True if the file path points to a file. False if it points to a directory.
        """
        return self.file_metadata(file_path).is_file

    def is_directory(self, file_path: str | Directory) -> bool:
        """
        Determine if the given file path points to a directory. This is achieved by checking the metadata of the
        provided file path. If the metadata is not cached, then this will make a webservice call to get the metadata.

        :param file_path: A file path or Directory object.
        :return: True if the file path points to a directory. False if it points to a file.
        """
        return self.file_metadata(file_path).is_directory

    def move_file(self, move_from: str | Directory, move_to: str | Directory, old_name: str | File,
                  new_name: str | File | None = None) -> None:
        """
        Move a file from one location to another within File Bridge. This is done be reading the file into memory,
        writing a copy of the file in the new location, then deleting the original file.

        :param move_from: The path or Directory object to the current location of the file.
        :param move_to: The path or Directory object to move the file to.
        :param old_name: The current name of the file, or a File object.
        :param new_name: The name that the file should have after it is moved, or a File object. If this is not
            provided, then the new name will be the same as the old name.
        """
        if isinstance(move_from, Directory):
            move_from = move_from.path
        if isinstance(move_to, Directory):
            move_to = move_to.path
        if isinstance(old_name, File):
            old_name = old_name.name
        if isinstance(new_name, File):
            new_name = new_name.name
        if not new_name:
            new_name = old_name

        # Read the file into memory.
        file_bytes: bytes = self.read_file(move_from + "/" + old_name)
        # Write the file into the new location.
        self.write_file(move_to + "/" + new_name, file_bytes)
        # Delete the file from the old location. We do this last in case the write call fails.
        self.delete_file(move_from + "/" + old_name)

    def get_file_object(self, file_path: str) -> File:
        """
        Get a File object from a file path. This object can be used to get the contents of the file at this path
        and traverse up the file hierarchy to the directory that the file is contained within.

        There is no guarantee that this file actually exists within the current file bridge connection when it is
        constructed. If the file doesn't exist, then retrieving its contents will fail.

        :param file_path: A file path.
        :return: A File object constructed form the given file path.
        """
        if file_path in self.__file_objects:
            return self.__file_objects[file_path]
        file = File(self, file_path)
        self.__file_objects[file_path] = file
        return file

    def get_directory_object(self, file_path: str) -> Directory:
        """
        Get a Directory object from a file path. This object can be used to traverse up and down the file hierarchy
        by going up to the parent directory that this directory is contained within or going down to the contents of
        this directory. A file path of "" (a blank string) equates to the root directory of this file bridge connection.

        There is no guarantee that this directory actually exists within the current file bridge connection when it is
        constructed. If the directory doesn't exist, then retrieving its contents will fail.

        :param file_path: A file path.
        :return: A Directory object constructed from the given file path.
        """
        if file_path in self.__dir_objects:
            return self.__dir_objects[file_path]
        directory = Directory(self, file_path)
        self.__dir_objects[file_path] = directory
        return directory


class FileBridgeObject(ABC):
    """
    A FileBridgeObject is either a file or a directory that is contained within file bridge. Every object has a
    name and a parent directory that it is contained within, unless the object is the root directory, in
    which case the parent is None. The root directory has a path and name of "" (a blank string).

    Note that this object may not actually exist within the file bridge connection that it is associated with.
    Retrieving the contents of an object that doesn't exist will fail. You can use the write_file or create_directory
    functions of the FileBridgeHandler to create new files or directories from these objects.
    """
    _handler: FileBridgeHandler
    _name: str
    _path_to: str
    _parent: Directory | None

    def __init__(self, handler: FileBridgeHandler, file_path: str):
        self._handler = handler

        # Remove any leading or trailing slashes from the file path.
        if file_path.startswith("/") or file_path.endswith("/"):
            file_path = file_path.strip("/")

        # If the file path is an empty string, then this is the root directory.
        if file_path == "":
            self._name = ""
            self._path_to = ""
            self._parent = None
            return
        name, path_to = split_path(file_path)
        self._name = name
        self._path_to = path_to
        self._parent = handler.get_directory_object(path_to)

    @property
    def name(self) -> str:
        """
        :return: The name of this object.
        """
        return self._name

    @property
    def path_to(self) -> str:
        """
        :return: The file path that leads to this object. Excludes the name of the object itself.
        """
        return self._path_to

    @property
    def path(self) -> str:
        """
        :return: The full file path that leads to this object. Includes the name of the object itself.
        """
        if self._path_to == "":
            return self._name
        return self._path_to + "/" + self._name

    @property
    def parent(self) -> Directory | None:
        """
        :return: The parent directory of this object. If this object is the root directory, then this will be None.
        """
        return self._parent

    @abstractmethod
    def is_file(self) -> bool:
        """
        :return: True if this object is a file. False if it is a directory.
        """
        pass

    @abstractmethod
    def is_directory(self) -> bool:
        """
        :return: True if this object is a directory. False if it is a file.
        """
        pass

    def exists(self) -> bool:
        """
        :return: True if this object exists in the file bridge connection that it is associated with.
            False if it does not.
        """
        return self._handler.file_exists(self.path)

    def get_metadata(self) -> FileBridgeMetadata:
        """
        :return: The metadata for this object.
        """
        return self._handler.file_metadata(self.path)


class File(FileBridgeObject):
    def __init__(self, handler: FileBridgeHandler, file_path: str):
        """
        :param handler: A FileBridgeHandler for the connection that this file came from.
        :param file_path: The path to this file.
        """
        super().__init__(handler, file_path)

    @property
    def contents(self) -> bytes:
        """
        Read the bytes of this file.
        This pulls from the cache of this object's related FileBridgeHandler.

        :return: The bytes of this file.
        """
        return self._handler.read_file(self.path)

    def is_file(self) -> bool:
        return True

    def is_directory(self) -> bool:
        return False


class Directory(FileBridgeObject):
    _contents: dict[str, FileBridgeObject] | None

    def __init__(self, handler: FileBridgeHandler, file_path: str):
        """
        :param handler: A FileBridgeHandler for the connection that this directory came from.
        :param file_path: The path to this directory.
        """
        super().__init__(handler, file_path)
        self._contents = None

    @property
    def contents(self) -> dict[str, FileBridgeObject]:
        """
        Get all the objects in this Directory.
        This pulls from the cache of this object's related FileBridgeHandler.

        :return: A dictionary of object names to the objects (Files or Directories) contained within this Directory.
        """
        if self._contents is not None:
            return self._contents

        # Load the metadata of the directory to get the names of the files and directories within it.
        # We don't need the return value of this function, but we need to call it to populate the cache.
        self._handler.directory_metadata(self._path_to)

        # Construct the objects for the contents of this directory.
        self._contents = {}
        for name in self._handler.list_directory(self._path_to):
            file_path: str = self._path_to + "/" + name
            if self._handler.is_file(file_path):
                self._contents[name] = self._handler.get_file_object(file_path)
            else:
                self._contents[name] = self._handler.get_directory_object(file_path)
        return self._contents

    def is_file(self) -> bool:
        return False

    def is_directory(self) -> bool:
        return True

    def get_files(self) -> dict[str, File]:
        """
        Get all the files in this Directory.
        This pulls from the cache of this object's related FileBridgeHandler.

        :return: A mapping of file name to File for every file in this Directory.
        """
        return {x: cast(File, y) for x, y in self.contents.items() if y.is_file()}

    def get_directories(self) -> dict[str, Directory]:
        """
        Get all the nested directories in this Directory.
        This pulls from the cache of this object's related FileBridgeHandler.

        :return: A mapping of directory name to Directory for every nested directory in this Directory.
        """
        return {x: cast(Directory, y) for x, y in self.contents.items() if not y.is_file()}


def split_path(file_path: str) -> tuple[str, str]:
    """
    :param file_path: A file path where directories are separated the "/" characters. If there is no "/" character, then
        that means that the provided path is in the root directory, or is the root directory itself.
    :return: A tuple of two strings that splits the path on its last slash. The first string is the name of the
        file/directory at the given file path and the second string is the location to that file. If there is no slash
        character, then the second string is an empty string.
    """
    last_slash: int = file_path.rfind("/")
    if last_slash == -1:
        return file_path, ""
    return file_path[last_slash + 1:], file_path[:last_slash]



import io

from sapiopylib.rest.User import SapioUser

from sapiopycommons.general.aliases import UserIdentifier, AliasUtil


class CDL:
    @staticmethod
    def load_cdl(context: UserIdentifier, config_name: str, file_name: str, file_data: bytes | str) \
            -> list[int]:
        """
        Create data records from a file using one of the complex data loader (CDL) configurations in the system.

        :param context: The current webhook context or a user object to send requests from.
        :param config_name: The name of the CDL configuration to run.
        :param file_name: The name of the file being read by the CDL.
        :param file_data: A string or bytes of the file to be read by the CDL.
        :return: A list of the record IDs of the data records created by the CDL.
        """
        sub_path = "/ext/cdl/load"
        params = {
            "configName": config_name,
            "fileName": file_name
        }
        user: SapioUser = AliasUtil.to_sapio_user(context)
        with io.BytesIO(file_data.encode() if isinstance(file_data, str) else file_data) as data_stream:
            response = user.post_data_stream(sub_path, params=params, data_stream=data_stream)
        user.raise_for_status(response)
        # The response content is returned as bytes for a comma separated string of record IDs.
        return [int(x) for x in bytes.decode(response.content).split(",")]



from sapiopylib.rest.utils.Protocols import ElnEntryStep
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType

from sapiopycommons.eln.experiment_handler import ExperimentHandler
from sapiopycommons.general.aliases import SapioRecord, RecordIdentifier, AliasUtil
from sapiopycommons.general.exceptions import SapioException
from sapiopycommons.recordmodel.record_handler import RecordHandler

PLATE_IDS_TAG: str = "MultiLayerPlating_Plate_RecordIdList"


class PlateDesignerEntry:
    """
    A wrapper for 3D plate designer entries in experiments, providing functions for common actions when dealing with
    such entries.
    """
    step: ElnEntryStep
    __exp_handler: ExperimentHandler
    __rec_handler: RecordHandler
    __plates: list[SapioRecord] | None
    __aliquots: list[SapioRecord] | None
    __sources: list[SapioRecord] | None
    __designer_elements: list[SapioRecord] | None
    __plate_ids: list[int] | None

    def __init__(self, step: ElnEntryStep, exp_handler: ExperimentHandler):
        """
        :param step: The ElnEntryStep that is the 3D plate designer entry.
        :param exp_handler: An ExperimentHandler for the experiment that this entry comes from.
        """
        self.step = step
        self.__exp_handler = exp_handler
        self.__rec_handler = RecordHandler(exp_handler.context)
        self.__plates = None
        self.__aliquots = None
        self.__sources = None
        self.__designer_elements = None
        self.__plate_ids = None

    def get_plates(self, wrapper_type: type[WrappedType]) -> list[WrappedType]:
        """
        Get the plates that are in the designer entry.

        Makes a webservice query to get the plates from the entry and caches the result for future calls. This cache
        will be invalidated if a set_plates or add_plates call is made, requiring a new webservice call the next time
        this function is called.

        :param wrapper_type: The record model wrapper to use.
        :return: A list of the plates in the designer entry.
        """
        if self.__plates is not None:
            return self.__plates
        self.__plates = self.__rec_handler.query_models_by_id(wrapper_type, self.__get_plate_ids())
        return self.__plates

    def set_plates(self, plates: list[RecordIdentifier]) -> None:
        """
        Set the plates that are in the plate designer entry. This removes any existing plates that are in the entry
        but not in the given list.

        Makes a webservice call to update the plate designer entry's entry options.

        :param plates: The plates to set the plate designer entry with.
        """
        record_ids: list[int] = AliasUtil.to_record_ids(plates)
        self.__set_plate_ids(record_ids)

    def add_plates(self, plates: list[RecordIdentifier]) -> None:
        """
        Add the given plates to the plate designer entry. This preserves any existing plates that are in the entry.

        Makes a webservice call to update the plate designer entry's entry options.

        :param plates: The plates to add to the plate designer entry.
        """
        record_ids: list[int] = AliasUtil.to_record_ids(plates)
        self.__set_plate_ids(self.__get_plate_ids() + record_ids)

    def get_sources(self, wrapper_type: type[WrappedType]) -> list[WrappedType]:
        """
        Get the source records that were used to populate the plate designer entry's sample table. This looks for the
        entries that the plate designer entry is dependent upon and gets their records if they match the data type name
        of the given wrapper.

        Makes a webservice call to retrieve the dependent entries if the experiment handler had not already cached it.
        Makes another webservice call to get the records from the dependent entry and caches them for future calls.

        :param wrapper_type: The record model wrapper to use.
        :return: A list of the source records that populate the plate designer entry's sample table.
        """
        if self.__sources is not None:
            return self.__sources

        records: list[WrappedType] = []
        dependent_ids: list[int] = self.step.eln_entry.dependency_set
        for step in self.__exp_handler.get_all_steps(wrapper_type):
            if step.get_id() in dependent_ids:
                records.extend(self.__exp_handler.get_step_models(step, wrapper_type))

        self.__sources = records
        return self.__sources

    def get_aliquots(self, wrapper_type: type[WrappedType]) -> list[WrappedType]:
        """
        Get the aliquots that were created from this plate designer entry upon its submission.

        Makes a webservice call to retrieve the aliquots from the plate designer entry and caches them for future calls.

        :param wrapper_type: The record model wrapper to use.
        :return: A list of the aliquots created by the plate designer entry.
        """
        if not self.__exp_handler.step_is_submitted(self.step):
            raise SapioException("The plate designer entry must be submitted before its aliquots can be retrieved.")
        if self.__aliquots is not None:
            return self.__aliquots
        self.__aliquots = self.__exp_handler.get_step_models(self.step, wrapper_type)
        return self.__aliquots

    def get_plate_designer_well_elements(self, wrapper_type: type[WrappedType]) -> list[WrappedType]:
        """
        Get the plate designer well elements for the plates in the plate designer entry. These are the records in the
        system that determine how wells are displayed on each plate in the entry.

        Makes a webservice call to get the plate designer well elements of the entry and caches them for future calls.
        This cache will be invalidated if a set_plates or add_plates call is made, requiring a new webservice call the
        next time this function is called.

        :param wrapper_type: The record model wrapper to use.
        :return: A list of the plate designer well elements in the designer entry.
        """
        if self.__designer_elements is not None:
            return self.__designer_elements
        self.__designer_elements = self.__rec_handler.query_models(wrapper_type, "PlateRecordId",
                                                                   self.__get_plate_ids())
        return self.__designer_elements

    def __get_plate_ids(self) -> list[int]:
        if self.__plate_ids is not None:
            return self.__plate_ids
        id_tag: str = self.__exp_handler.get_step_option(self.step, PLATE_IDS_TAG)
        if not id_tag:
            raise SapioException("No plates in the plate designer entry")
        self.__plate_ids = [int(x) for x in id_tag.split(",")]
        return self.__plate_ids

    def __set_plate_ids(self, record_ids: list[int]) -> None:
        record_ids.sort()
        self.__exp_handler.add_step_options(self.step, {PLATE_IDS_TAG: ",".join([str(x) for x in record_ids])})
        self.__plate_ids = record_ids
        # The plates and designer elements caches have been invalidated.
        self.__plates = None
        self.__designer_elements = None



from indigo import Indigo, IndigoObject
from indigo.inchi import IndigoInchi
from indigo.renderer import IndigoRenderer

indigo = Indigo()
renderer = IndigoRenderer(indigo)
indigo.setOption("render-output-format", "svg")
indigo.setOption("ignore-stereochemistry-errors", True)
indigo.setOption("render-stereo-style", "ext")
indigo.setOption("aromaticity-model", "generic")
indigo.setOption("render-coloring", True)
indigo.setOption("molfile-saving-mode", "3000")
indigo_inchi = IndigoInchi(indigo)


# Function to process dative bonds in a molecule
# Returns True if at least one dative bond (_BOND_COORDINATION) was removed
def remove_dative_bonds_in_mol(molecule: IndigoObject) -> bool:
    """
    Remove all dative bonds in a molecule or a query molecule.
    :param molecule: The molecule to remove.
    :return: Whether there are any dative bonds in the molecule that were removed.
    """
    dative_bond_removed = False  # Flag to track if any dative bond was removed

    bonds_idx_to_remove = []
    for bond in molecule.iterateBonds():
        # Check if the bond is of a dative type (_BOND_COORDINATION = 9)
        if bond.bondOrder() == 9:  # _BOND_COORDINATION
            atom1 = bond.source()
            atom2 = bond.destination()

            # Print bond details for debugging
            print(f"Processing dative bond between atoms {atom1.index()} and {atom2.index()}")

            # Cache bond information
            # bond.setBondOrder(1)
            bonds_idx_to_remove.append(bond.index())
            dative_bond_removed = True  # Set flag to True

    if not dative_bond_removed:
        return False

    molecule.removeBonds(bonds_idx_to_remove)
    return True  # Return whether any dative bond was removed


def remove_dative_in_reaction(reaction: IndigoObject) -> bool:
    """
    Remove all dative bonds in a reaction or a query reaction, from all reactants and products.
    :param reaction: The reaction to remove dative bonds.
    :return: Whether there are any dative bonds in the reaction that were removed.
    """
    reactant_dative_removed: bool = any(remove_dative_bonds_in_mol(reactant) for reactant in reaction.iterateReactants())
    product_dative_removed: bool = any(remove_dative_bonds_in_mol(product) for product in reaction.iterateProducts())
    return reactant_dative_removed or product_dative_removed


def highlight_mol_substructure(query: IndigoObject, sub_match: IndigoObject):
    """
    Highlight the bonds and atoms for substructure search result.

    :param sub_match: The substructure search match obtained from indigo.substructureMatcher(mol).match(query).
    :param query: The query we were running to match the original structure.
    """
    for qatom in query.iterateAtoms():
        atom = sub_match.mapAtom(qatom)
        if atom is None:
            continue
        atom.highlight()

        for nei in atom.iterateNeighbors():
            if not nei.isPseudoatom() and not nei.isRSite() and nei.atomicNumber() == 1:
                nei.highlight()
                nei.bond().highlight()

    for bond in query.iterateBonds():
        bond = sub_match.mapBond(bond)
        if bond is None:
            continue
        bond.highlight()


def highlight_reactions(query_reaction_smarts: IndigoObject, reaction_match: IndigoObject):
    """
    Highlight the bonds and atoms for substructure search result of reaction that's in the query and survived the mapping.

    :param query_reaction_smarts: The query we ran substructure search on.
    :param reaction_match: The substructure search match obtained from indigo.substructureMatcher(reaction).match(query).
    :return:
    """
    for q_mol in query_reaction_smarts.iterateMolecules():
        matched_mol = reaction_match.mapMolecule(q_mol)
        sub_match = indigo.substructureMatcher(matched_mol).match(q_mol)
        highlight_mol_substructure(q_mol, sub_match)



import json
from typing import Any

from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext


class CustomWebhookContext(SapioWebhookContext):
    """
    CustomWebhookContext is a wrapper for the default SapioWebhookContext to be used by custom invocation types to
    convert the context_data that the server sends from a JSON string to usable parameters. This class works as a
    middleman to cleanly convert a SapioWebhookContext object into a CustomWebhookContext by copying all the parameters
    of the given context into the custom context.
    """
    def __init__(self, context: SapioWebhookContext):
        # Use __dict__ so that we don't need t maintain this class due to future changes to SapioWebhookContext.
        self.__dict__ = context.__dict__
        super().__init__(self.user, self.end_point_type)


class ProcessQueueContext(CustomWebhookContext):
    """
    When a custom process queue endpoint is invoked, the context from the queue is sent in a context_data parameter
    on the SapioWebhookContext object, stored as a JSON string. This class parses that JSON into fields for the
    caller to make use of.
    """
    process_name: str
    """The name of the process that the user invoked this webhook from."""
    step_name: str
    """The name of the step in the process that the user invoked this webhook from."""
    process_queue_item_record_ids: list[int]
    """The record IDs of the process queue items related to the records that were selected by the user when this
    webhook was invoked."""

    def __init__(self, context: SapioWebhookContext):
        super().__init__(context)
        context_data: dict[str, Any] = json.loads(self.context_data)
        self.process_name = context_data["processName"]
        self.step_name = context_data["workflowName"]
        self.process_queue_item_record_ids = context_data["processQueueItemRecordIds"]



from sapiopylib.rest.pojo.datatype.FieldDefinition import FieldType
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrapperField


class ActiveTaskPseudoDef:
    DATA_TYPE_NAME: str = "ActiveTask"
    ACTIVE_TASK_ID__FIELD_NAME = WrapperField("ActiveTaskId", FieldType.LONG)
    ACTIVE_WORKFLOW_ID__FIELD_NAME = WrapperField("ActiveWorkflowId", FieldType.LONG)
    DATE_EDITED__FIELD_NAME = WrapperField("DateEdited", FieldType.DATE)
    EDITED_BY__FIELD_NAME = WrapperField("EditedBy", FieldType.STRING)
    STATUS__FIELD_NAME = WrapperField("Status", FieldType.ENUM)
    TASK_USAGE_ID__FIELD_NAME = WrapperField("TaskUsageId", FieldType.LONG)


class ActiveWorkflowPseudoDef:
    DATA_TYPE_NAME: str = "ActiveWorkflow"
    ACTIVE_WORKFLOW_ID__FIELD_NAME = WrapperField("ActiveWorkflowId", FieldType.LONG)
    CREATED_BY__FIELD_NAME = WrapperField("CreatedBy", FieldType.STRING)
    DATE_CREATED__FIELD_NAME = WrapperField("DateCreated", FieldType.DATE)
    DATE_EDITED__FIELD_NAME = WrapperField("DateEdited", FieldType.STRING)
    EDITED_BY__FIELD_NAME = WrapperField("EditedBy", FieldType.STRING)
    ESTIMATED_ATTACHMENTS__FIELD_NAME = WrapperField("EstimatedAttachments", FieldType.LONG)
    NAME__FIELD_NAME = WrapperField("Name", FieldType.STRING)
    RELATED_RECORD_ID__FIELD_NAME = WrapperField("RelatedRecordId", FieldType.LONG)
    STATUS__FIELD_NAME = WrapperField("Status", FieldType.ENUM)
    WORKFLOW_ID__FIELD_NAME = WrapperField("WorkflowId", FieldType.LONG)


class AuditLogPseudoDef:
    DATA_TYPE_NAME: str = "AuditLog"
    DATA_FIELD_NAME__FIELD_NAME = WrapperField("DataFieldName", FieldType.STRING)
    DATA_TYPE_NAME__FIELD_NAME = WrapperField("DataTypeName", FieldType.STRING)
    DESCRIPTION__FIELD_NAME = WrapperField("Description", FieldType.STRING)
    EVENT_TYPE__FIELD_NAME = WrapperField("EventType", FieldType.ENUM)
    FULL_NAME__FIELD_NAME = WrapperField("FullName", FieldType.STRING)
    NEW_VALUE__FIELD_NAME = WrapperField("NewValue", FieldType.STRING)
    ORIGINAL_VALUE__FIELD_NAME = WrapperField("OriginalValue", FieldType.STRING)
    RECORD_ID__FIELD_NAME = WrapperField("RecordId", FieldType.LONG)
    RECORD_NAME__FIELD_NAME = WrapperField("RecordName", FieldType.STRING)
    TIME_STAMP__FIELD_NAME = WrapperField("TimeStamp", FieldType.DATE)
    USER_COMMENT__FIELD_NAME = WrapperField("UserComment", FieldType.STRING)
    USER_NAME__FIELD_NAME = WrapperField("UserName", FieldType.STRING)


class DataFieldDefinitionPseudoDef:
    DATA_TYPE_NAME: str = "DataFieldDefinition"
    APPROVE_EDIT__FIELD_NAME = WrapperField("ApproveEdit", FieldType.BOOLEAN)
    AUTO_CLEAR_FIELD_LIST__FIELD_NAME = WrapperField("AutoClearFieldList", FieldType.STRING)
    AUTO_SORT__FIELD_NAME = WrapperField("AutoSort", FieldType.BOOLEAN)
    COLOR_MAPPING_ID__FIELD_NAME = WrapperField("ColorMappingId", FieldType.LONG)
    DATA_FIELD_NAME__FIELD_NAME = WrapperField("DataFieldName", FieldType.STRING)
    DATA_FIELD_TAG__FIELD_NAME = WrapperField("DataFieldTag", FieldType.STRING)
    DATA_FIELD_TYPE__FIELD_NAME = WrapperField("DataFieldType", FieldType.STRING)
    DATA_TYPE_NAME__FIELD_NAME = WrapperField("DataTypeName", FieldType.STRING)
    DECIMAL_DIGITS__FIELD_NAME = WrapperField("DecimalDigits", FieldType.INTEGER)
    DEFAULT_VALUE__FIELD_NAME = WrapperField("DefaultValue", FieldType.STRING)
    DEPENDENT_FIELDS__FIELD_NAME = WrapperField("Dependent_Fields", FieldType.STRING)
    DESCRIPTION__FIELD_NAME = WrapperField("Description", FieldType.STRING)
    DIRECT_EDIT__FIELD_NAME = WrapperField("DirectEdit", FieldType.BOOLEAN)
    DISPLAY_NAME__FIELD_NAME = WrapperField("DisplayName", FieldType.STRING)
    EDITABLE__FIELD_NAME = WrapperField("Editable", FieldType.BOOLEAN)
    ENUM_VALUES__FIELD_NAME = WrapperField("EnumValues", FieldType.STRING)
    FORM_COL__FIELD_NAME = WrapperField("FormCol", FieldType.SHORT)
    FORM_COL_SPAN__FIELD_NAME = WrapperField("FormColSpan", FieldType.SHORT)
    FORM_NAME__FIELD_NAME = WrapperField("FormName", FieldType.STRING)
    HTML_EDITOR__FIELD_NAME = WrapperField("HtmlEditor", FieldType.BOOLEAN)
    IDENTIFIER__FIELD_NAME = WrapperField("Identifier", FieldType.BOOLEAN)
    INDEX_FOR_SEARCH__FIELD_NAME = WrapperField("IndexForSearch", FieldType.BOOLEAN)
    LINK_OUT__FIELD_NAME = WrapperField("LinkOut", FieldType.BOOLEAN)
    LINK_OUT_URL__FIELD_NAME = WrapperField("LinkOutUrl", FieldType.BOOLEAN)
    MAX_LENGTH__FIELD_NAME = WrapperField("MaxLength", FieldType.INTEGER)
    MAXIMUM_VALUE__FIELD_NAME = WrapperField("MaximumValue", FieldType.DOUBLE)
    MINIMUM_VALUE__FIELD_NAME = WrapperField("MinimumValue", FieldType.DOUBLE)
    MULT_SELECT__FIELD_NAME = WrapperField("MultSelect", FieldType.BOOLEAN)
    NUM_LINES__FIELD_NAME = WrapperField("NumLines", FieldType.INTEGER)
    REQUIRED__FIELD_NAME = WrapperField("Required", FieldType.BOOLEAN)
    SORT_KEY__FIELD_NAME = WrapperField("SortKey", FieldType.BOOLEAN)
    STATIC_DATE__FIELD_NAME = WrapperField("StaticDate", FieldType.BOOLEAN)
    UNIQUE_VALUE__FIELD_NAME = WrapperField("UniqueValue", FieldType.BOOLEAN)
    VISIBLE__FIELD_NAME = WrapperField("Visible", FieldType.BOOLEAN)
    WORKFLOW_ONLY_EDITING__FIELD_NAME = WrapperField("WorkflowOnlyEditing", FieldType.BOOLEAN)


class DataTypeDefinitionPseudoDef:
    DATA_TYPE_NAME: str = "DataTypeDefinition"
    ADDABLE__FIELD_NAME = WrapperField("Addable", FieldType.BOOLEAN)
    ATTACHMENT__FIELD_NAME = WrapperField("Attachment", FieldType.BOOLEAN)
    ATTACHMENT_TYPE__FIELD_NAME = WrapperField("AttachmentType", FieldType.STRING)
    DATA_TYPE_TAG__FIELD_NAME = WrapperField("DATA_TYPE_TAG", FieldType.STRING)
    DATA_TYPE_ID__FIELD_NAME = WrapperField("DataTypeId", FieldType.LONG)
    DATA_TYPE_NAME__FIELD_NAME = WrapperField("DataTypeName", FieldType.STRING)
    DELETABLE__FIELD_NAME = WrapperField("Deletable", FieldType.BOOLEAN)
    DESCRIPTION__FIELD_NAME = WrapperField("Description", FieldType.STRING)
    DISPLAY_NAME__FIELD_NAME = WrapperField("DisplayName", FieldType.STRING)
    EXTENSION_TYPE__FIELD_NAME = WrapperField("ExtensionType", FieldType.BOOLEAN)
    GROUP_ADDABLE__FIELD_NAME = WrapperField("GroupAddable", FieldType.BOOLEAN)
    HIDE_DATA_RECORDS__FIELD_NAME = WrapperField("HideDataRecords", FieldType.BOOLEAN)
    HIGH_VOLUME__FIELD_NAME = WrapperField("HighVolume", FieldType.BOOLEAN)
    IS_HIDE_IN_MOBILE__FIELD_NAME = WrapperField("IS_HIDE_IN_MOBILE", FieldType.BOOLEAN)
    IS_HVDT_ON_SAVE_ENABLED__FIELD_NAME = WrapperField("IS_HVDT_ON_SAVE_ENABLED", FieldType.BOOLEAN)
    IS_PUBLIC_ATTACHMENT__FIELD_NAME = WrapperField("IS_PUBLIC_ATTACHMENT", FieldType.BOOLEAN)
    ICON_COLOR__FIELD_NAME = WrapperField("IconColor", FieldType.STRING)
    ICON_NAME__FIELD_NAME = WrapperField("IconName", FieldType.STRING)
    IMPORTABLE__FIELD_NAME = WrapperField("Importable", FieldType.BOOLEAN)
    IS_ACTIVE__FIELD_NAME = WrapperField("IsActive", FieldType.BOOLEAN)
    IS_HIDDEN__FIELD_NAME = WrapperField("IsHidden", FieldType.BOOLEAN)
    MAX_TABLE_ROW_COUNT__FIELD_NAME = WrapperField("MaxTableRowCount", FieldType.LONG)
    PLURAL_DISPLAY_NAME__FIELD_NAME = WrapperField("PluralDisplayName", FieldType.STRING)
    RECORD_ASSIGNABLE__FIELD_NAME = WrapperField("RecordAssignable", FieldType.BOOLEAN)
    RECORD_IMAGE_ASSIGNABLE__FIELD_NAME = WrapperField("RecordImageAssignable", FieldType.BOOLEAN)
    RECORD_IMAGE_MANUALLY_ADDABLE__FIELD_NAME = WrapperField("RecordImageManuallyAddable", FieldType.BOOLEAN)
    REMOVABLE__FIELD_NAME = WrapperField("Removable", FieldType.BOOLEAN)
    RESTRICTED__FIELD_NAME = WrapperField("Restricted", FieldType.BOOLEAN)
    SHOW_ON_HOME_SCREEN__FIELD_NAME = WrapperField("ShowOnHomeScreen", FieldType.BOOLEAN)
    SHOW_SUB_TABLES__FIELD_NAME = WrapperField("ShowSubTables", FieldType.BOOLEAN)
    SHOW_TABS__FIELD_NAME = WrapperField("ShowTabs", FieldType.BOOLEAN)
    SINGLE_PARENT__FIELD_NAME = WrapperField("SingleParent", FieldType.BOOLEAN)
    UNDER_CONTAINER__FIELD_NAME = WrapperField("UnderContainer", FieldType.BOOLEAN)


class EnbDataTypeDefinitionPseudoDef:
    DATA_TYPE_NAME: str = "EnbDataTypeDefinition"
    DATA_TYPE_ID__FIELD_NAME = WrapperField("DataTypeId", FieldType.LONG)
    DATA_TYPE_NAME__FIELD_NAME = WrapperField("DataTypeName", FieldType.STRING)
    DESCRIPTION__FIELD_NAME = WrapperField("Description", FieldType.STRING)
    DISPLAY_NAME__FIELD_NAME = WrapperField("DisplayName", FieldType.STRING)
    ENB_DATA_TYPE_NAME__FIELD_NAME = WrapperField("EnbDataTypeName", FieldType.STRING)
    ICON_COLOR__FIELD_NAME = WrapperField("IconColor", FieldType.STRING)
    ICON_NAME__FIELD_NAME = WrapperField("IconName", FieldType.STRING)
    NOTEBOOK_EXPERIMENT_ID__FIELD_NAME = WrapperField("Notebook_Experiment_ID", FieldType.LONG)
    PLURAL_DISPLAY_NAME__FIELD_NAME = WrapperField("PluralDisplayName", FieldType.STRING)


class EnbEntryPseudoDef:
    DATA_TYPE_NAME: str = "EnbEntry"
    APPROVAL_DUE_DATE__FIELD_NAME = WrapperField("ApprovalDueDate", FieldType.DATE)
    COLUMN_ORDER__FIELD_NAME = WrapperField("ColumnOrder", FieldType.INTEGER)
    COLUMN_SPAN__FIELD_NAME = WrapperField("ColumnSpan", FieldType.INTEGER)
    CREATED_BY__FIELD_NAME = WrapperField("CreatedBy", FieldType.STRING)
    DATA_TYPE_NAME__FIELD_NAME = WrapperField("DataTypeName", FieldType.STRING)
    DATE_CREATED__FIELD_NAME = WrapperField("DateCreated", FieldType.DATE)
    DEPENDENT_ENTRY_ID_LIST__FIELD_NAME = WrapperField("DependentEntryIdList", FieldType.STRING)
    ENTRY_DESCRIPTION__FIELD_NAME = WrapperField("EntryDescription", FieldType.STRING)
    ENTRY_HEIGHT__FIELD_NAME = WrapperField("EntryHeight", FieldType.INTEGER)
    ENTRY_ID__FIELD_NAME = WrapperField("EntryId", FieldType.LONG)
    ENTRY_NAME__FIELD_NAME = WrapperField("EntryName", FieldType.STRING)
    ENTRY_ORDER__FIELD_NAME = WrapperField("EntryOrder", FieldType.INTEGER)
    ENTRY_REQUIRES_E_SIGN__FIELD_NAME = WrapperField("EntryRequiresESign", FieldType.BOOLEAN)
    ENTRY_SINGLETON_ID__FIELD_NAME = WrapperField("EntrySingletonId", FieldType.STRING)
    ENTRY_STATUS__FIELD_NAME = WrapperField("EntryStatus", FieldType.STRING)
    ENTRY_TYPE__FIELD_NAME = WrapperField("EntryType", FieldType.STRING)
    EXPERIMENT_ID__FIELD_NAME = WrapperField("ExperimentId", FieldType.LONG)
    HAS_COMMENTS__FIELD_NAME = WrapperField("HasComments", FieldType.BOOLEAN)
    IS_CREATED_FROM_TEMPLATE__FIELD_NAME = WrapperField("IsCreatedFromTemplate", FieldType.BOOLEAN)
    IS_REQUIRED_ENTRY__FIELD_NAME = WrapperField("IsRequiredEntry", FieldType.BOOLEAN)
    IS_SHOWN_IN_TEMPLATE__FIELD_NAME = WrapperField("IsShownInTemplate", FieldType.BOOLEAN)
    LAST_MODIFIED_BY__FIELD_NAME = WrapperField("LastModifiedBy", FieldType.STRING)
    LAST_MODIFIED_DATE__FIELD_NAME = WrapperField("LastModifiedDate", FieldType.DATE)
    RELATED_ENTRY_ID_LIST__FIELD_NAME = WrapperField("RelatedEntryIdList", FieldType.STRING)
    REQUIRES_GRABBER_PLUGIN__FIELD_NAME = WrapperField("RequriesGrabberPlugin", FieldType.BOOLEAN)
    SOURCE_ENTRY_ID__FIELD_NAME = WrapperField("SourceEntryId", FieldType.LONG)
    SUBMITTED_BY__FIELD_NAME = WrapperField("SubmittedBy", FieldType.STRING)
    SUBMITTED_DATE__FIELD_NAME = WrapperField("SubmittedDate", FieldType.DATE)
    TAB_ID__FIELD_NAME = WrapperField("TabId", FieldType.LONG)
    TEMPLATE_ITEM_FULFILLED_TIME_STAMP__FIELD_NAME = WrapperField("TemplateItemFulfilledTimeStamp", FieldType.LONG)


class EnbEntryOptionsPseudoDef:
    DATA_TYPE_NAME: str = "EnbEntryOptions"
    ENTRY_ID__FIELD_NAME = WrapperField("EntryId", FieldType.LONG)
    ENTRY_OPTION_VALUE__FIELD_NAME = WrapperField("EntryOptionValue", FieldType.STRING)
    ENTRY_OPTION_KEY__FIELD_NAME = WrapperField("EntryOptionkey", FieldType.STRING)


class ExperimentEntryRecordPseudoDef:
    DATA_TYPE_NAME: str = "ExperimentEntryRecord"
    ENTRY_ID__FIELD_NAME = WrapperField("EntryId", FieldType.LONG)
    RECORD_ID__FIELD_NAME = WrapperField("RecordId", FieldType.LONG)


class LimsUserPseudoDef:
    DATA_TYPE_NAME: str = "LimsUser"
    API_USER__FIELD_NAME = WrapperField("ApiUser", FieldType.BOOLEAN)
    EMAIL_ADDRESS__FIELD_NAME = WrapperField("EmailAddress", FieldType.STRING)
    FIRST_NAME__FIELD_NAME = WrapperField("FirstName", FieldType.STRING)
    LAST_NAME__FIELD_NAME = WrapperField("LastName", FieldType.STRING)
    MIDDLE_NAME__FIELD_NAME = WrapperField("MiddleName", FieldType.STRING)
    PWD_EXPIRE_DATE__FIELD_NAME = WrapperField("PwdExpireDate", FieldType.DATE)
    PWD_EXPIRE_INTERVAL__FIELD_NAME = WrapperField("PwdExpireInterval", FieldType.INTEGER)
    USER_NAME__FIELD_NAME = WrapperField("UserName", FieldType.STRING)


class NotebookExperimentPseudoDef:
    DATA_TYPE_NAME: str = "NotebookExperiment"
    ACCESS_LEVEL__FIELD_NAME = WrapperField("AccessLevel", FieldType.STRING)
    APPROVAL_DUE_DATE__FIELD_NAME = WrapperField("ApprovalDueDate", FieldType.DATE)
    CREATED_BY__FIELD_NAME = WrapperField("CreatedBy", FieldType.STRING)
    DATE_CREATED__FIELD_NAME = WrapperField("DateCreated", FieldType.DATE)
    DESCRIPTION__FIELD_NAME = WrapperField("Description", FieldType.STRING)
    EXPERIMENT_ID__FIELD_NAME = WrapperField("ExperimentId", FieldType.LONG)
    EXPERIMENT_NAME__FIELD_NAME = WrapperField("ExperimentName", FieldType.STRING)
    EXPERIMENT_OWNER__FIELD_NAME = WrapperField("ExperimentOwner", FieldType.STRING)
    EXPERIMENT_RECORD_ID__FIELD_NAME = WrapperField("ExperimentRecordId", FieldType.LONG)
    EXPERIMENT_TYPE_NAME__FIELD_NAME = WrapperField("ExperimentTypeName", FieldType.STRING)
    IS_ACTIVE__FIELD_NAME = WrapperField("IsActive", FieldType.BOOLEAN)
    IS_MODIFIABLE__FIELD_NAME = WrapperField("IsModifiable", FieldType.BOOLEAN)
    IS_TEMPLATE__FIELD_NAME = WrapperField("IsTemplate", FieldType.BOOLEAN)
    IS_PROTOCOL_TEMPLATE__FIELD_NAME = WrapperField("Is_Protocol_Template", FieldType.BOOLEAN)
    LAST_MODIFIED_BY__FIELD_NAME = WrapperField("LastModifiedBy", FieldType.STRING)
    LAST_MODIFIED_DATE__FIELD_NAME = WrapperField("LastModifiedDate", FieldType.DATE)
    SOURCE_TEMPLATE_ID__FIELD_NAME = WrapperField("SourceTemplateId", FieldType.LONG)
    STATUS__FIELD_NAME = WrapperField("Status", FieldType.STRING)


class NotebookExperimentOptionPseudoDef:
    DATA_TYPE_NAME: str = "NotebookExperimentOption"
    EXPERIMENT_ID__FIELD_NAME = WrapperField("ExperimentId", FieldType.LONG)
    OPTION_KEY__FIELD_NAME = WrapperField("OptionKey", FieldType.STRING)
    OPTION_VALUE__FIELD_NAME = WrapperField("OptionValue", FieldType.STRING)


class SystemLogPseudoDef:
    DATA_TYPE_NAME: str = "SystemLog"
    DATA_FIELD_NAME__FIELD_NAME = WrapperField("DataFieldName", FieldType.STRING)
    DATA_TYPE_NAME__FIELD_NAME = WrapperField("DataTypeName", FieldType.STRING)
    DESCRIPTION__FIELD_NAME = WrapperField("Description", FieldType.STRING)
    EVENT_ID__FIELD_NAME = WrapperField("EventId", FieldType.LONG)
    EVENT_TYPE__FIELD_NAME = WrapperField("EventType", FieldType.STRING)
    FULL_NAME__FIELD_NAME = WrapperField("FullName", FieldType.STRING)
    NEW_VALUE__FIELD_NAME = WrapperField("NewValue", FieldType.STRING)
    ORIGINAL_VALUE__FIELD_NAME = WrapperField("OriginalValue", FieldType.STRING)
    RECORD_ID__FIELD_NAME = WrapperField("RecordId", FieldType.LONG)
    RECORD_NAME__FIELD_NAME = WrapperField("RecordName", FieldType.STRING)
    TIMESTAMP__FIELD_NAME = WrapperField("Timestamp", FieldType.DATE)
    USER_COMMENT__FIELD_NAME = WrapperField("UserComment", FieldType.STRING)
    USERNAME__FIELD_NAME = WrapperField("Username", FieldType.STRING)


class SystemObjectChangeLogPseudoDef:
    DATA_TYPE_NAME: str = "System_Object_Change_Log"
    ALT_ID__FIELD_NAME = WrapperField("Alt_Id", FieldType.STRING)
    ATTRIBUTE_NAME__FIELD_NAME = WrapperField("Attribute_Name", FieldType.STRING)
    CHANGE_TYPE__FIELD_NAME = WrapperField("Change_Type", FieldType.STRING)
    DATA_FIELD_NAME__FIELD_NAME = WrapperField("Data_Field_Name", FieldType.STRING)
    DATA_TYPE_NAME__FIELD_NAME = WrapperField("Data_Type_Name", FieldType.STRING)
    EVENT_ID__FIELD_NAME = WrapperField("Event_Id", FieldType.STRING)
    NEW_VALUE__FIELD_NAME = WrapperField("New_Value", FieldType.STRING)
    OBJECT_ID__FIELD_NAME = WrapperField("Object_Id", FieldType.STRING)
    OBJECT_TYPE__FIELD_NAME = WrapperField("Object_Type", FieldType.STRING)
    OLD_VALUE__FIELD_NAME = WrapperField("Old_Value", FieldType.STRING)
    TIMESTAMP__FIELD_NAME = WrapperField("Timestamp", FieldType.DATE)
    USERNAME__FIELD_NAME = WrapperField("Username", FieldType.STRING)


class TaskPseudoDef:
    DATA_TYPE_NAME: str = "Task"
    ATTACHMENT_REQUIRED__FIELD_NAME = WrapperField("AttachmentRequired", FieldType.BOOLEAN)
    CREATED_BY__FIELD_NAME = WrapperField("CreatedBy", FieldType.STRING)
    CUSTOM_ACTION__FIELD_NAME = WrapperField("CustomAction", FieldType.STRING)
    DATA_TYPE_NAME_LIST__FIELD_NAME = WrapperField("DataTypeNameList", FieldType.STRING)
    DATE_CREATED__FIELD_NAME = WrapperField("DateCreated", FieldType.DATE)
    DATE_EDITED__FIELD_NAME = WrapperField("DateEdited", FieldType.DATE)
    DISPLAY_TYPE__FIELD_NAME = WrapperField("DisplayType", FieldType.ENUM)
    EDITED_BY__FIELD_NAME = WrapperField("EditedBy", FieldType.STRING)
    INPUT_DATA_TYPE_NAME__FIELD_NAME = WrapperField("InputDataTypeName", FieldType.STRING)
    IS_TEMPLATE__FIELD_NAME = WrapperField("IsTemplate", FieldType.BOOLEAN)
    LONG_DESC__FIELD_NAME = WrapperField("LongDesc", FieldType.STRING)
    MENU_TASK_ID__FIELD_NAME = WrapperField("MenuTaskId", FieldType.ENUM)
    NAME__FIELD_NAME = WrapperField("Name", FieldType.STRING)
    SHORT_DESC__FIELD_NAME = WrapperField("ShortDesc", FieldType.STRING)
    TASK_ID__FIELD_NAME = WrapperField("TaskId", FieldType.LONG)
    TASK_VERSION__FIELD_NAME = WrapperField("TaskVersion", FieldType.LONG)
    TEMPLATE_TASK_ID__FIELD_NAME = WrapperField("TemplateTaskId", FieldType.LONG)
    TYPE__FIELD_NAME = WrapperField("Type", FieldType.ENUM)


class TaskAttachmentPseudoDef:
    DATA_TYPE_NAME: str = "TaskAttachment"
    ACTIVE_TASK_ID__FIELD_NAME = WrapperField("ActiveTaskId", FieldType.LONG)
    RECORD_ID__FIELD_NAME = WrapperField("RecordId", FieldType.LONG)


class TaskOptionPseudoDef:
    DATA_TYPE_NAME: str = "TaskOption"
    OPTION_KEY__FIELD_NAME = WrapperField("OptionKey", FieldType.STRING)
    OPTION_VALUE__FIELD_NAME = WrapperField("OptionValue", FieldType.STRING)
    TASK_ID__FIELD_NAME = WrapperField("TaskId", FieldType.LONG)


class TaskUsagePseudoDef:
    DATA_TYPE_NAME: str = "TaskUsage"
    FORCE_ATTACH__FIELD_NAME = WrapperField("ForceAttach", FieldType.BOOLEAN)
    IS_TEMPLATE__FIELD_NAME = WrapperField("IsTemplate", FieldType.BOOLEAN)
    TASK_ID__FIELD_NAME = WrapperField("TaskId", FieldType.LONG)
    TASK_ORDER__FIELD_NAME = WrapperField("TaskOrder", FieldType.INTEGER)
    TASK_USAGE_ID__FIELD_NAME = WrapperField("TaskUsageId", FieldType.LONG)
    WORKFLOW_ID__FIELD_NAME = WrapperField("WorkflowId", FieldType.LONG)


class VeloxWebhookPseudoDef:
    DATA_TYPE_NAME: str = "VELOXWEBHOOK"
    CUSTOM_PLUGIN_POINT__FIELD_NAME = WrapperField("CUSTOM_PLUGIN_POINT", FieldType.STRING)
    DATA_TYPE_NAME_SET__FIELD_NAME = WrapperField("DATA_TYPE_NAME_SET", FieldType.STRING)
    DESCRIPTION__FIELD_NAME = WrapperField("DESCRIPTION", FieldType.STRING)
    ENB_ENTRY_TYPE__FIELD_NAME = WrapperField("ENB_ENTRY_TYPE", FieldType.STRING)
    EXPERIMENT_ENTRY_NAME_SET__FIELD_NAME = WrapperField("EXPERIMENT_ENTRY_NAME_SET", FieldType.STRING)
    GUID__FIELD_NAME = WrapperField("GUID", FieldType.STRING)
    ICON_COLOR__FIELD_NAME = WrapperField("ICON_COLOR", FieldType.STRING)
    ICON_GUID__FIELD_NAME = WrapperField("ICON_GUID", FieldType.STRING)
    IS_RETRY_ON_FAILURE__FIELD_NAME = WrapperField("IS_RETRY_ON_FAILURE", FieldType.BOOLEAN)
    IS_TRANSACTIONAL__FIELD_NAME = WrapperField("IS_TRANSACTIONAL", FieldType.BOOLEAN)
    LINE_1_TEXT__FIELD_NAME = WrapperField("LINE_1_TEXT", FieldType.STRING)
    LINE_2_TEXT__FIELD_NAME = WrapperField("LINE_2_TEXT", FieldType.STRING)
    PLUGIN_ORDER__FIELD_NAME = WrapperField("PLUGIN_ORDER", FieldType.INTEGER)
    PLUGIN_POINT__FIELD_NAME = WrapperField("PLUGIN_POINT", FieldType.STRING)
    SECTION_NAME_PATH__FIELD_NAME = WrapperField("SECTION_NAME_PATH", FieldType.STRING)
    TEMPLATE_NAME__FIELD_NAME = WrapperField("TEMPLATE_NAME", FieldType.STRING)
    WEBHOOK_URL__FIELD_NAME = WrapperField("WEBHOOK_URL", FieldType.STRING)


class VeloxWebhookExecutionPseudoDef:
    DATA_TYPE_NAME: str = "VELOXWEBHOOK_EXECUTION"
    EXECUTION_TIMESTAMP__FIELD_NAME = WrapperField("EXECUTION_TIMESTAMP", FieldType.DATE)
    EXECUTION_USERNAME__FIELD_NAME = WrapperField("EXECUTION_USERNAME", FieldType.STRING)
    GUID__FIELD_NAME = WrapperField("GUID", FieldType.STRING)
    LAST_ATTEMPT_NUMBER__FIELD_NAME = WrapperField("LAST_ATTEMPT_NUMBER", FieldType.INTEGER)
    LAST_ATTEMPT_RESULT__FIELD_NAME = WrapperField("LAST_ATTEMPT_RESULT", FieldType.STRING)
    REQUEST_BODY__FIELD_NAME = WrapperField("REQUEST_BODY", FieldType.STRING)
    WEBHOOK_GUID__FIELD_NAME = WrapperField("WEBHOOK_GUID", FieldType.STRING)
    WEBHOOK_URL__FIELD_NAME = WrapperField("WEBHOOK_URL", FieldType.STRING)


class VeloxWebhookExecutionAttemptPseudoDef:
    DATA_TYPE_NAME: str = "VELOXWEBHOOK_EXECUTION_ATTEMPT"
    ATTEMPT_DURATION__FIELD_NAME = WrapperField("ATTEMPT_DURATION", FieldType.INTEGER)
    ATTEMPT_NUMBER__FIELD_NAME = WrapperField("ATTEMPT_NUMBER", FieldType.INTEGER)
    ATTEMPT_RESULT__FIELD_NAME = WrapperField("ATTEMPT_RESULT", FieldType.STRING)
    ATTEMPT_TIMESTAMP__FIELD_NAME = WrapperField("ATTEMPT_TIMESTAMP", FieldType.DATE)
    EXECUTION_GUID__FIELD_NAME = WrapperField("EXECUTION_GUID", FieldType.STRING)
    GUID__FIELD_NAME = WrapperField("GUID", FieldType.STRING)
    RESPONSE_BODY__FIELD_NAME = WrapperField("RESPONSE_BODY", FieldType.STRING)
    RESPONSE_CODE__FIELD_NAME = WrapperField("RESPONSE_CODE", FieldType.INTEGER)
    WEBHOOK_GUID__FIELD_NAME = WrapperField("WEBHOOK_GUID", FieldType.STRING)
    WEBHOOK_URL__FIELD_NAME = WrapperField("WEBHOOK_URL", FieldType.STRING)


class VeloxWebhookExecutionLogPseudoDef:
    DATA_TYPE_NAME: str = "VELOXWEBHOOK_EXECUTION_LOG"
    ATTEMPT_GUID__FIELD_NAME = WrapperField("ATTEMPT_GUID", FieldType.STRING)
    LOG_LEVEL__FIELD_NAME = WrapperField("LOG_LEVEL", FieldType.STRING)
    LOG_LINE_NUM__FIELD_NAME = WrapperField("LOG_LINE_NUM", FieldType.INTEGER)
    LOG_MESSAGE__FIELD_NAME = WrapperField("LOG_MESSAGE", FieldType.STRING)
    LOG_TIMESTAMP__FIELD_NAME = WrapperField("LOG_TIMESTAMP", FieldType.DATE)


class VeloxRuleCostPseudoDef:
    DATA_TYPE_NAME: str = "VELOX_RULE_COST"
    ACTION_COST__FIELD_NAME = WrapperField("ACTION_COST", FieldType.LONG)
    ACTION_COUNT__FIELD_NAME = WrapperField("ACTION_COUNT", FieldType.LONG)
    ANCESTOR_DESCENDANT_COUNT__FIELD_NAME = WrapperField("ANCESTOR_DESCENDANT_COUNT", FieldType.LONG)
    PARENT_CHILD_COUNT__FIELD_NAME = WrapperField("PARENT_CHILD_COUNT", FieldType.LONG)
    PROCESSING_TIME__FIELD_NAME = WrapperField("PROCESSING_TIME", FieldType.LONG)
    RULE_GUID__FIELD_NAME = WrapperField("RULE_GUID", FieldType.STRING)
    SOURCE_RECORD_COUNT__FIELD_NAME = WrapperField("SOURCE_RECORD_COUNT", FieldType.LONG)
    TIMESTAMP__FIELD_NAME = WrapperField("TIMESTAMP", FieldType.DATE)
    TOTAL_COST__FIELD_NAME = WrapperField("TOTAL_COST", FieldType.LONG)
    TRANSACTION_GUID__FIELD_NAME = WrapperField("TRANSACTION_GUID", FieldType.STRING)
    USERNAME__FIELD_NAME = WrapperField("USERNAME", FieldType.STRING)


class VeloxConversationPseudoDef:
    DATA_TYPE_NAME: str = "VeloxConversation"
    CONVERSATION_DESCRIPTION__FIELD_NAME = WrapperField("ConversationDescription", FieldType.STRING)
    CONVERSATION_GUID__FIELD_NAME = WrapperField("ConversationGuid", FieldType.STRING)
    CONVERSATION_NAME__FIELD_NAME = WrapperField("ConversationName", FieldType.STRING)
    CREATED_BY__FIELD_NAME = WrapperField("CreatedBy", FieldType.STRING)
    DATE_CREATED__FIELD_NAME = WrapperField("DateCreated", FieldType.DATE)
    SERVER_PLUGIN_PATH__FIELD_NAME = WrapperField("Server_Plugin_Path", FieldType.STRING)


class VeloxConversationMessagePseudoDef:
    DATA_TYPE_NAME: str = "VeloxConversationMessage"
    CONVERSATION_GUID__FIELD_NAME = WrapperField("ConversationGuid", FieldType.STRING)
    MESSAGE__FIELD_NAME = WrapperField("Message", FieldType.STRING)
    MESSAGE_GUID__FIELD_NAME = WrapperField("MessageGuid", FieldType.STRING)
    MESSAGE_SENDER__FIELD_NAME = WrapperField("MessageSender", FieldType.STRING)
    MESSAGE_TIMESTAMP__FIELD_NAME = WrapperField("MessageTimestamp", FieldType.DATE)


class VeloxScriptPseudoDef:
    DATA_TYPE_NAME: str = "VeloxScript"
    CODE__FIELD_NAME = WrapperField("Code", FieldType.STRING)
    CREATED_BY__FIELD_NAME = WrapperField("CreatedBy", FieldType.STRING)
    DATE_CREATED__FIELD_NAME = WrapperField("DateCreated", FieldType.LONG)
    LAST_MODIFIED_BY__FIELD_NAME = WrapperField("LastModifiedBy", FieldType.STRING)
    LAST_MODIFIED_DATE__FIELD_NAME = WrapperField("LastModifiedDate", FieldType.LONG)
    PATH__FIELD_NAME = WrapperField("Path", FieldType.STRING)
    PLUGIN_DESCRIPTION__FIELD_NAME = WrapperField("PluginDescription", FieldType.STRING)
    PLUGIN_LINE1_TEXT__FIELD_NAME = WrapperField("PluginLine1Text", FieldType.STRING)
    PLUGIN_LINE2_TEXT__FIELD_NAME = WrapperField("PluginLine2Text", FieldType.STRING)
    PLUGIN_ORDER__FIELD_NAME = WrapperField("PluginOrder", FieldType.INTEGER)
    PLUGIN_POINT__FIELD_NAME = WrapperField("PluginPoint", FieldType.STRING)
    PROJECT_GUID__FIELD_NAME = WrapperField("ProjectGuid", FieldType.STRING)
    SCRIPT_GUID__FIELD_NAME = WrapperField("ScriptGuid", FieldType.STRING)


class VeloxScriptProjectPseudoDef:
    DATA_TYPE_NAME: str = "VeloxScriptProject"
    CLASS_PATH__FIELD_NAME = WrapperField("ClassPath", FieldType.STRING)
    CREATED_BY__FIELD_NAME = WrapperField("CreatedBy", FieldType.STRING)
    DATE_CREATED__FIELD_NAME = WrapperField("DateCreated", FieldType.LONG)
    DEPLOYMENT_OUT_OF_DATE__FIELD_NAME = WrapperField("DeploymentOutOfDate", FieldType.BOOLEAN)
    DESCRIPTION__FIELD_NAME = WrapperField("Description", FieldType.STRING)
    PROJECT_GUID__FIELD_NAME = WrapperField("ProjectGuid", FieldType.STRING)
    PROJECT_NAME__FIELD_NAME = WrapperField("ProjectName", FieldType.STRING)
    SCRIPT_LANGUAGE__FIELD_NAME = WrapperField("ScriptLanguage", FieldType.STRING)


class WorkflowPseudoDef:
    DATA_TYPE_NAME: str = "Workflow"
    ALL_ACCESS__FIELD_NAME = WrapperField("AllAccess", FieldType.BOOLEAN)
    CATEGORY__FIELD_NAME = WrapperField("Category", FieldType.STRING)
    CREATED_BY__FIELD_NAME = WrapperField("CreatedBy", FieldType.STRING)
    DATE_CREATED__FIELD_NAME = WrapperField("DateCreated", FieldType.DATE)
    DATE_EDITED__FIELD_NAME = WrapperField("DateEdited", FieldType.DATE)
    DIRECT_LAUNCH__FIELD_NAME = WrapperField("DirectLaunch", FieldType.BOOLEAN)
    EDITED_BY__FIELD_NAME = WrapperField("EditedBy", FieldType.STRING)
    IS_TEMPLATE__FIELD_NAME = WrapperField("IsTemplate", FieldType.BOOLEAN)
    LONG_DESC__FIELD_NAME = WrapperField("LongDesc", FieldType.STRING)
    NAME__FIELD_NAME = WrapperField("Name", FieldType.STRING)
    SHORT_DESC__FIELD_NAME = WrapperField("ShortDesc", FieldType.STRING)
    WORKFLOW_ID__FIELD_NAME = WrapperField("WorkflowId", FieldType.LONG)
    WORKFLOW_VERSION__FIELD_NAME = WrapperField("WorkflowVersion", FieldType.LONG)


class WorkflowOptionPseudoDef:
    DATA_TYPE_NAME: str = "WorkflowOption"
    OPTION_KEY__FIELD_NAME = WrapperField("OptionKey", FieldType.STRING)
    OPTION_VALUE__FIELD_NAME = WrapperField("OptionValue", FieldType.STRING)
    WORKFLOW_ID__FIELD_NAME = WrapperField("WorkflowId", FieldType.LONG)
